/* process.c     process and communication part of the interpreter
 *
 *
 *  copyright(c)  Prabhas Chongstitvatana  28 Feb 97
 *
 *  add timer interrupt with the help of Aj. Chatchawal 17 Aug 98
 *  add xTmWait                                         21 Aug 98
*/

#include "interp.h"
#include <dos.h>

#define   PRIVATE	static

#define   semvalue(s)   M[s]
#define   semnext(s)    M[(s)+1]
#define   execprim(c)	(*fntab[c])()

extern fnptr fntab[];
extern uchar cs[];

pdata pcb[maxprocess];  /* process control block */
pdesc activep;		/* current active process */

int pclock;
int flagtimeout;
PRIVATE void interrupt( *oldhandler)();


PRIVATE void interrupt timeout()
{
  setvect(0x1C, oldhandler);
  /* do my task */
  flagtimeout = 1;
  setvect(0x1C, timeout);
  oldhandler();   		/* service old interrupt */
}

void restore_isr()		/* restore old isr */
{
  setvect(0x1C,oldhandler);
}

pdesc initpcb( p )
pdesc p;
{
  pcb[p].pid = p;
  pcb[p].prev = nil;
  pcb[p].next = nil;
  pcb[p].state = INIT;
  pcb[p].ip = 0;
  pcb[p].fp = 0;
  pcb[p].sp = (p-1)*stacksize - 1 + DSSIZE;
  pcb[p].maxsp = pcb[p].sp + stacksize;
  pcb[p].await = nil;
  pcb[p].mbox = nil;
  return p;
}

/* insert p at the left of list (the end) */
PRIVATE insertp( list, p )
pdesc *list,p;
{
  pdesc m,pv;
  m = *list;
  if( m == nil ) m = p;
  pv = pcb[m].prev;
  pcb[p].prev = pv;
  pcb[m].prev = p;
  pcb[p].next = m;
  pcb[pv].next = p;
  *list = m;
}

/* delete the first one in the list */
PRIVATE deletep( list )
pdesc *list;
{
  pdesc m,p,n;
  m = *list;
  if( m == nil ) return;
  p = pcb[m].prev;
  n = pcb[m].next;
  pcb[p].next = n;
  pcb[n].prev = p;
  pcb[m].prev = nil;
  pcb[m].next = nil;
  if( m == n ) *list = nil;  /* delete the last one in the list */
  else *list = n;
}

/* make p run-able as the only process in the task list by
   updating C-state with p's ip, sp, fp
*/
PRIVATE void runable( pdesc p )
{
  ip = pcb[p].ip;
  sp = pcb[p].sp;
  fp = pcb[p].fp;
  maxsp = pcb[p].maxsp;
  pcb[p].state = RUNNING;
}

void wakeup( pdesc p )
{
  pcb[p].state = READY;
  if( activep == nil ) {
    insertp(&activep, p);
    runable(p);
  } else {
    insertp(&activep, p);
  }
}

/* move p to the end of ready list */
void awakep( pdesc p )
{
  pstate s;
  s = pcb[p].state;
  if( s == TMWAIT || s == TMSEND || s == TMRECEIVE )
    deletetmid(p);
  wakeup(p);
}

/* block the current active process */
blockp( pstate status )
{
  pcb[activep].state = status;
  pcb[activep].ip = ip;		/* save C-state */
  pcb[activep].sp = sp;
  pcb[activep].fp = fp;
  deletep(&activep);
  if( activep == nil ) return;	/* restart next process */
  runable(activep);
}

/* switch to the next process */
switchp()
{
  vclock++;
  dotimer();

  if ( activep == nil ) return;
  if ( activep != pcb[activep].next ) {
    pcb[activep].ip = ip;
    pcb[activep].sp = sp;
    pcb[activep].fp = fp;
    pcb[activep].state = READY;
    activep = pcb[activep].next;
    runable(activep);
  }
}

/*  create the first process running the main routine */
void startup()
{
  ip = 1;
  sp = DSSIZE - 1;
  fp = 0;
  execprim(cs[ip]);  	/* call main */
  execprim(cs[ip]);  	/* create main process */
  runable(1);
  activep = 1;		/* start main */
  flagtimeout = 0;
  vclock = 0;
  pclock = 0;
#ifndef DEBUG
  oldhandler = getvect(0x1C);
  setvect(0x1C,timeout);
#endif
}

/* create one process control block and its initial stack frame
   awake it and return to the caller
 */
xProc()
{
  pdesc p;
  int i,a,sp1,fp1;
  p = initpcb( op1() );
  sp1 = pcb[p].sp + 1;
  a = op2();
  for( i = 0; i < a; i++ )
    M[sp1+i] = M[sp-a+i];
  fp1 = sp1 + a + op3();
  M[fp1+2] = 0;
  M[fp1+1] = 0;
  M[fp1] = 0;
  pcb[p].ip = ip + 7;
  pcb[p].fp = fp1;
  pcb[p].sp = fp1 + 2;
  awakep(p);
  ip = TOS;            /* caller's return address at TOS */
  sp = sp - a - 1;     /* adjust caller's stack */
}

#ifdef DEBUG
showp(p)
pdesc p;
{
  if ( p == nil ) {printf("nil\n"); return; }
  printf("pid %d prev %d next %d state %d\n",pcb[p].pid, pcb[p].prev,
    pcb[p].next, pcb[p].state);
  printf("ip %d fp %d sp %d maxsp %d\n",pcb[p].ip, pcb[p].fp, pcb[p].sp,
    pcb[p].maxsp);
  printf("await %d mbox %d\n",pcb[p].await, pcb[p].mbox);
}

showall()
{
  pdesc p;
  p = activep;
  while(1) {
    showp(p);
    p = pcb[p].next;
    if ( p == activep ) break;
  }
  printf(":\n");
}
#endif

/* ------------ semaphore operations -------------------- */

/* insert pcb of p to waitlist of a semaphore */
PRIVATE insertsem( s, p )
sem s;
pdesc p;
{
  pcb[p].next = semnext(s);
  semnext(s) = p;
}

/* remove the head of the list */
PRIVATE pdesc removesem( s )
sem s;
{
  pdesc p;
  p = semnext(s);
  semnext(s) = pcb[p].next;
  pcb[p].next = nil;
  return p;
}

void removesemid( sem s, pdesc p )
{
  pdesc a, b;
  a = semnext(s);
  b = a;
  while( a != nil && pcb[a].pid != p ) {
    b = a;
    a = pcb[a].next;
  }
  if( a == nil )
    error("attempt to remove non exist process from semaphore wait");
  if( a == b ) /* at the head of list */
    semnext(s) = pcb[a].next;
  else
    pcb[b].next = pcb[a].next;
}

xWait()
{
  pdesc p;
  sem s;
  s = op1();
  ip += 3;
  if ( semvalue(s) <= 0 ) {
    p = activep;
    blockp(WAIT);
    insertsem(s,p);
  } else
    semvalue(s)--;
}

xTmWait()
{
  pdesc p;
  sem s;
  int t;
  s = op1();
  ip += 3;
  t = pop();
  if ( semvalue(s) <= 0 ) {
    starttimer(t,s);
    p = activep;
    blockp(TMWAIT);
    insertsem(s,p);
  } else
    semvalue(s)--;
}

xSignal()
{
  pdesc p;
  sem s;
  s = op1();
  ip += 3;
  if( semnext(s) != nil ) {
    p = removesem(s);
    awakep(p);
  } else
    semvalue(s)++;
}

